/*
   Copyright (C) 2001  Ulric Eriksson <ulric@siag.nu>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the Licence, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.
 */

/* ---
Spinner.c - Spinner compound widget
--- */

#include	<stdio.h>
#include	<stdlib.h>
#include	<X11/keysym.h>
#include	<X11/IntrinsicP.h>
#include	<X11/StringDefs.h>
#include	<X11/Shell.h>
#include	<X11/Xaw/Repeater.h>
#include	<X11/Xmu/Misc.h>
#include	<X11/xpm.h>
#include	"FrameP.h"
#include	"Rudegrid.h"
#include	"TextField.h"
#include	"SpinnerP.h"
#include	"spinup.xpm"
#include	"spindown.xpm"

/****************************************************************
 *
 * Spinner Resources
 *
 ****************************************************************/

#define XtNspinnerShell "spinnerShell"
#define XtCSpinnerShell "SpinnerShell"

#define offset(field) XtOffsetOf(SpinnerRec, spinner.field)
static XtResource resources[] = {
	{
		XtNmin,		/* name */
		XtCMin,		/* class */
		XtRInt,		/* type */
		sizeof(int),	/* size */
		offset(min),	/* offset */
		XtRImmediate,	/* default_type */
		(XtPointer)0	/* default_addr */
	}, {
		XtNmax,
		XtCMax,
		XtRInt,
		sizeof(int),
		offset(max),
		XtRImmediate,
		(XtPointer)100
	}, {
		XtNstep,
		XtCStep,
		XtRInt,
		sizeof(int),
		offset(step),
		XtRImmediate,
		(XtPointer)1
	}, {
		XtNvalue,
		XtCValue,
		XtRInt,
		sizeof(int),
		offset(value),
		XtRImmediate,
		(XtPointer)0
	},{
		XtNcallback,
		XtCCallback,
		XtRCallback,
		sizeof(XtPointer),
		offset(callbacks),
		XtRCallback,
		(XtPointer)NULL
	},{
		XtNshadowType,
		XtCShadowType,
		XtRShadowType,
		sizeof(XtShadowType),
		XtOffsetOf(FrameRec, frame.type),
		XtRImmediate,
		(XtPointer)Lowered
	},{
		XtNshadowWidth,
		XtCShadowWidth,
		XtRDimension,
		sizeof(Dimension),
		XtOffsetOf(FrameRec, frame.shadowWidth),
		XtRImmediate,
		(XtPointer)1
	}
};
#undef offset

/****************************************************************
 *
 * Full class record constant
 *
 ****************************************************************/


static void Initialize(Widget, Widget, ArgList, Cardinal *);
static void Destroy(Widget);

static Boolean SetValues(Widget, Widget, Widget, ArgList, Cardinal *);

SpinnerClassRec spinnerClassRec = {
  {
/* core_class fields      */
    /* superclass         */    (WidgetClass) &frameClassRec,
    /* class_name         */    "Spinner",
    /* widget_size        */    sizeof(SpinnerRec),
    /* class_initialize   */    NULL,
    /* class_part_init    */	NULL,
    /* class_inited       */	FALSE,
    /* initialize	  */	Initialize,
    /* initialize_hook    */	NULL,
    /* realize            */    XtInheritRealize,
    /* actions            */    NULL,
    /* num_actions	  */	0,
    /* resources          */    resources,
    /* num_resources      */    XtNumber(resources),
    /* xrm_class          */    NULLQUARK,
    /* compress_motion	  */	TRUE,
    /* compress_exposure  */	TRUE,
    /* compress_enterleave*/	TRUE,
    /* visible_interest   */    FALSE,
    /* destroy            */    Destroy,
    /* resize             */    XtInheritResize,
    /* expose             */    XtInheritExpose,
    /* set_values         */    SetValues,
    /* set_values_hook    */	NULL,
    /* set_values_almost  */    XtInheritSetValuesAlmost,
    /* get_values_hook    */	NULL,
    /* accept_focus       */    NULL,
    /* version            */	XtVersion,
    /* callback_private   */    NULL,
    /* tm_table           */    NULL,
    /* query_geometry     */	XtInheritQueryGeometry,
    /* display_accelerator*/	XtInheritDisplayAccelerator,
    /* extension          */	NULL
  },{
/* composite_class fields */
    /* geometry_manager   */    XtInheritGeometryManager,
    /* change_managed     */    XtInheritChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension          */	NULL
  },{
/* Frame class fields */
    /* extension	  */	NULL
  },{
/* Spinner class fields */
    /* empty		  */	0
  }
};

WidgetClass spinnerWidgetClass = (WidgetClass)&spinnerClassRec;


/****************************************************************
 *
 * Private Routines
 *
 ****************************************************************/

static enum {WAITING, ABORT, DONE} spinner_status;

static void set_text(SpinnerWidget sw)
{
	char b[100];
	int n = sw->spinner.value;
	if (n > sw->spinner.max) n = sw->spinner.max;
	if (n < sw->spinner.min) n = sw->spinner.min;
	sw->spinner.value = n;
	sprintf(b, "%d", n);
	TextFieldSetString(sw->spinner.text, b);
}

static void do_callback(SpinnerWidget sw)
{
	XtCallCallbackList((Widget)sw, sw->spinner.callbacks, (XtPointer)NULL);
}

/* ---
This event handler is triggered when the user clicks in the text widget.
   It allows editing the text until Esc or Return is pressed.
   Pressing Return sets the widget value to the entered text and
   calls callbacks. Pressing Esc returns the text to what it was before.
*/

static void spinner_text_edit(Widget w, XtPointer p, XEvent *event, Boolean *n)
{
	SpinnerWidget cw = (SpinnerWidget)XtParent(XtParent(w));
	Widget top;

	for (top = w; XtParent(top) != None; top = XtParent(top));
	XtAddGrab(w, True, False);
	XtSetKeyboardFocus(top, w);

	XtVaSetValues(w,
		XtNdisplayCaret, True,
		(char *)0);
	spinner_status = WAITING;
	while (spinner_status == WAITING) {
		int count, bufsiz = 10;
		XEvent event;
		XtAppNextEvent(XtWidgetToApplicationContext(w), &event);
		if (event.type == KeyPress) {
			char buf[12];
			KeySym keysym;
			XKeyEvent kevent = event.xkey;
			count = XLookupString(&kevent,
				buf, bufsiz, &keysym, NULL);
			if (keysym == XK_Escape) spinner_status = ABORT;
			else if (keysym == XK_Return) spinner_status = DONE;
			else XtDispatchEvent(&event);
		} else {
			XtDispatchEvent(&event);
		}
	}
	XtVaSetValues(w,
		XtNdisplayCaret, False,
		(char *)0);
	XtRemoveGrab(w);
	XtSetKeyboardFocus(top, None);

	if (spinner_status == DONE) {
		char *b = TextFieldGetString(w);
		cw->spinner.value = atoi(b);
		set_text(cw);
		do_callback(cw);
	} else {
		set_text(cw);
	}
}

static void spinup_cb(Widget w, XtPointer p, XtPointer q)
{
	SpinnerWidget sw = (SpinnerWidget)XtParent(XtParent(w));
	sw->spinner.value += sw->spinner.step;
	set_text(sw);
	do_callback(sw);
}

static void spindown_cb(Widget w, XtPointer p, XtPointer q)
{
	SpinnerWidget sw = (SpinnerWidget)XtParent(XtParent(w));
	sw->spinner.value -= sw->spinner.step;
	set_text(sw);
	do_callback(sw);
}

static void Initialize(Widget req, Widget new, ArgList args, Cardinal *num)
{
	XpmAttributes xa;
	Pixmap pm_return;
	int result;
	XpmColorSymbol symbol;
	Pixel color;
	SpinnerWidget cw = (SpinnerWidget)new;

	XtVaGetValues(new,
		XtNbackground, &color,
		(char *)0);
	xa.closeness = 40000;
	xa.exactColors = False;
	xa.valuemask = XpmCloseness | XpmExactColors;
	symbol.name = NULL;
	symbol.value = "none";
	symbol.pixel = color;
	xa.colorsymbols = &symbol;
	xa.numsymbols = 1;
	xa.valuemask |= XpmColorSymbols;

	cw->spinner.grid = XtVaCreateManagedWidget("spinner_grid",
		rudegridWidgetClass, new,
		XtNxLayout, "100% 16",
		XtNyLayout, "50% 50%",
		XtNwidth, 100,
		XtNheight, 20,
		XtNborderWidth, 0,
		(char *)0);
	cw->spinner.text = XtVaCreateManagedWidget("spinner_text",
		textfieldWidgetClass, cw->spinner.grid,
		XtNdisplayCaret, False,
		XtNgridHeight, 2,
		(char *)0);
	XtAddEventHandler(cw->spinner.text,
		ButtonPressMask, False, spinner_text_edit, NULL);
	set_text(cw);

	result = XpmCreatePixmapFromData(XtDisplay(new),
#if 1
		XRootWindowOfScreen(XtScreen(new)),
#else
		XtWindow(new),
#endif
		spinup_xpm, &pm_return, NULL, &xa);
	if (result != XpmSuccess) {
		fprintf(stderr, "XpmCreatePixmapFromData returns %s\n",
			XpmGetErrorString(result));
	}
	cw->spinner.up = XtVaCreateManagedWidget("spinner_up",
		repeaterWidgetClass, cw->spinner.grid,
		XtNbitmap, pm_return,
		XtNforeground, color,
		XtNgridx, 1,
		(char *)0);
	XtAddCallback(cw->spinner.up, XtNcallback, spinup_cb, NULL);

	result = XpmCreatePixmapFromData(XtDisplay(new),
		XRootWindowOfScreen(XtScreen(new)),
		spindown_xpm, &pm_return, NULL, &xa);
	if (result != XpmSuccess) {
		fprintf(stderr, "XpmCreatePixmapFromData returns %s\n",
			XpmGetErrorString(result));
	}
	cw->spinner.down = XtVaCreateManagedWidget("spinner_down",
		repeaterWidgetClass, cw->spinner.grid,
		XtNbitmap, pm_return,
		XtNforeground, color,
		XtNgridx, 1,
		XtNgridy, 1,
		(char *)0);
	XtAddCallback(cw->spinner.down, XtNcallback, spindown_cb, NULL);
}

static void Destroy(Widget w)
{
	SpinnerWidget cw = (SpinnerWidget)w;
	XtDestroyWidget(cw->spinner.up);
	XtDestroyWidget(cw->spinner.down);
	XtDestroyWidget(cw->spinner.text);
	XtDestroyWidget(cw->spinner.grid);
}

static Boolean SetValues(Widget current, Widget request, Widget new,
		ArgList args, Cardinal *num_args)
{
	set_text((SpinnerWidget)new);
    return False;
}

